"""
Settings for django-facebook
"""
import facebook
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

try:
    APP_ID = getattr(settings, 'FACEBOOK_APP_ID')
    APP_SECRET = getattr(settings, 'FACEBOOK_APP_SECRET')
    REDIRECT_URI = getattr(settings, 'FACEBOOK_REDIRECT_URI')
except AttributeError:
    raise ImproperlyConfigured('You need to set FACEBOOK_APP_ID, '
        'FACEBOOK_APP_SECRET and FACEBOOK_REDIRECT_URI to use django-facebook')

VERSION = getattr(settings, 'FACEBOOK_VERSION', "2.2")

auth = facebook.Auth(APP_ID, APP_SECRET, REDIRECT_URI, VERSION)
COOKIE_NAME = 'fbsr_%s' % APP_ID
CANVAS_PAGE = getattr(settings, 'FACEBOOK_CANVAS_PAGE', "")
DEBUG_SIGNEDREQ = getattr(settings, 'FACEBOOK_DEBUG_SIGNEDREQ', "")
DEBUG_COOKIE = getattr(settings, 'FACEBOOK_DEBUG_COOKIE', "")
DEBUG_TOKEN = getattr(settings, 'FACEBOOK_DEBUG_TOKEN', "")



import hashlib
import logging

from django.contrib import auth as django_auth
from django.contrib.auth import BACKEND_SESSION_KEY, get_user_model, SESSION_KEY
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.signals import user_logged_in
from django_facebook.signals import facebook_user_created
from django_facebook.utils import (cache_access_token, del_cached_access_token,
                                   del_cached_fb_user_data,
                                   get_signed_request_data)

import conf

User = get_user_model()

log = logging.getLogger('django_facebook.auth')


def login(request, user):
    """
    Persist the facebook user_id and the backend in the request. This way a
    user doesn't have to reauthenticate on every request.
    """
    if user is None:
        user = request.user
    if SESSION_KEY in request.session:
        if request.session[SESSION_KEY] != user.get_username():
            # To avoid reusing another user's session, create a new, empty
            # session if the existing session corresponds to a different
            # authenticated user.
            request.session.flush()
    else:
        request.session.cycle_key()
    # save the username (is facebook id) as user id
    request.session[SESSION_KEY] = user.get_username()
    request.session[BACKEND_SESSION_KEY] = user.backend
    if hasattr(request, 'user'):
        request.user = user
    if hasattr(request, 'facebook'):
        request.facebook.user_id = user.get_username()

    log.debug('Facebook user %s logged in' % user.get_username())
    user_logged_in.send(sender=user.__class__, request=request, user=user)


def logout(request):
    """
    Logout the user, delete cached data and clear cookies so any auth calls
    coming after don't log the user in again.
    """
    user_id = request.user.get_username()
    del_cached_fb_user_data(user_id)
    del_cached_access_token(user_id)
    django_auth.logout(request)

    try:
        del request.COOKIES[conf.COOKIE_NAME]
    except KeyError:
        pass


class FacebookModelBackend(ModelBackend):

    create_on_not_found = True

    def authenticate(self, request=None, access_token=None, expires_in=None,
                     force_validate=False):
        """
        Users are basically authenticated by the signed_request facebook gives
        us, as facebook takes care of the real authentication. You must
        therefore only call this method with a request, upon which the ``fbsr_``
        cookie is used to validate the user, or with the ``signed_request``
        itself.

        If you pass in the ``signed_request``, you must also provide the
        ``access_token`` and ``expires_in`` kwargs.
        """
        if not request:
            raise TypeError("Please provide the request")

        user = None
        if request:
            auth_data = get_signed_request_data(request)
            if auth_data:
                user = self.get_user(auth_data['user_id'], access_token)

                if access_token:
                    cache_access_token(auth_data['user_id'],
                                       access_token,
                                       expires_in)
                    log.debug('access_token cached for user %s' % auth_data['user_id'])

        return user

    def get_user(self, user_id, access_token=None):
        """
        Lookup the user by their facebook id, and create a new one if they
        don't exist yet. Upon creation the facebook_user_created signal is
        fired. For this reason you really should pass a valid access_token for
        this user, so that connecting functions can actually do something
        usefull, like pre-fetching user data.
        """
        log.debug('FacebookModelBackend.get_user called')
        if self.create_on_not_found:
            user, created = User.objects.get_or_create(
                **{User.USERNAME_FIELD: user_id,
                   'defaults': {'password': '!'}})  # also set unusable password
            if created:
                log.debug('New user created for facebook account %s' % user_id)
                kwargs = dict(sender=self, user=user, access_token=access_token)
                facebook_user_created.send_robust(**kwargs)
            return user
        else:
            return User.objects.filter(**{User.USERNAME_FIELD: user_id}).first()



import facebook
from django.utils.functional import SimpleLazyObject
from django.contrib.auth import BACKEND_SESSION_KEY
from django.core.cache import cache

import conf

FB_ACCESS_TOKEN_CACHE_KEY = '_fb_access_token_%s'
FB_DATA_CACHE_KEY = '_fb_data_%s'


def get_lazy_access_token(request):
    if request.user.is_anonymous():
        return None
    code, use_redirect_uri = get_code_from_request(request)
    fb_id = request.user.get_username()

    def get_lazy():
        access_token = get_cached_access_token(fb_id)

        if not access_token:
            try:
                access_token, expires_in = get_fresh_access_token(code, use_redirect_uri)
                cache_access_token(fb_id, access_token, expires_in)
            except facebook.AuthError:
                raise

        return access_token

    return SimpleLazyObject(get_lazy)


def get_code_from_request(request):
    """
    Fetches the code from either the GET params or the signed_request cookie.

    Returns the code and wether or not a redirect_uri should be used.
    """
    code = None
    use_redirect_uri = True
    if 'code' in request.GET:
        code = request.GET['code']
        use_redirect_uri = True
    elif 'code' in get_signed_request_data(request):
        code = get_signed_request_data(request)['code']
        use_redirect_uri = False

    return code, use_redirect_uri


def get_fresh_access_token(code, use_redirect_uri=True):
    """
    Get a fresh_access_token with the provided code. Raise a
    facebook.AuthError if we can't.

    If the user is logged in through the client side, use_redirect_uri should
    be False.

    Returns the access_token and the amount of seconds it expires in.
    """
    if not code:
        raise facebook.AuthError('OAuthException', 'Their is no code to get an'
            'access_token with. Reauthenticate the user')

    try:
        # freaking facebook doesn't want a redirect_uri is somebody is logged
        # in throught the client-side...
        kwargs = {'redirect_uri': ''} if not use_redirect_uri else {}
        data = conf.auth.get_access_token_from_code(code, **kwargs)
    except facebook.AuthError:
        raise

    return data['access_token'], data['expires']


def get_signed_request_data(request):
    """
    Cache parsed signed_request cookie data, so we only do it once per
    request.
    """
    if not hasattr(request, '_fb_cookie_data'):
        try:
            data = conf.auth.parse_signed_request(request.COOKIES[conf.COOKIE_NAME])
        except (KeyError, ValueError, facebook.AuthError):
            data = {}
        request._fb_cookie_data = data
    return request._fb_cookie_data


def is_fb_logged_in(request):
    backendstr = 'django_facebook.auth.FacebookModelBackend'
    return request.user.is_authenticated() and \
        request.session.get(BACKEND_SESSION_KEY) == backendstr


class FacebookRequiredMixin(object):
    """CBV view mixin to display a facebook login template when the user is
    not logged in.
    """
    facebook_required_template = "facebook_required.html"

    def dispatch(self, request, *args, **kwargs):
        if not is_fb_logged_in(request):
            self.template_name = self.facebook_required_template
            self.request = request
            self.object = None
            return self.render_to_response({})
        return super(FacebookRequiredMixin, self).dispatch(request, *args, **kwargs)


def cache_access_token(user_id, access_token, expires_in=3600):
    """Cache the access_token for a user"""
    cache.set(FB_ACCESS_TOKEN_CACHE_KEY % user_id, access_token, int(expires_in))


def get_cached_access_token(user_id, default=None):
    return cache.get(FB_ACCESS_TOKEN_CACHE_KEY % user_id, default)


def del_cached_access_token(user_id):
    cache.delete(FB_ACCESS_TOKEN_CACHE_KEY % user_id)


def cache_fb_user_data(user_id, data, expires_in=None):
    if not expires_in:
        cache.set(FB_DATA_CACHE_KEY % user_id, data)
    else:
        cache.set(FB_DATA_CACHE_KEY % user_id, data, int(expires_in))


def del_cached_fb_user_data(user_id):
    cache.delete(FB_DATA_CACHE_KEY % user_id)


def get_cached_fb_user_data(user_id, default=None):
    return cache.get(FB_DATA_CACHE_KEY % user_id, default)



import urllib
import logging

from django.core.urlresolvers import reverse
from django.contrib.auth import logout, authenticate
from django.http import (HttpResponse, HttpResponseRedirect,
    HttpResponseNotAllowed, HttpResponseBadRequest)
from django.conf import settings
from django.template.loader import render_to_string
from django.views.decorators.csrf import csrf_exempt

import facebook

import conf
from .auth import login, FacebookModelBackend
from .utils import cache_access_token, is_fb_logged_in


log = logging.getLogger('django_facebook.views')


def fb_server_login(request):
    """View that accepts the redirect from Facebook after the user signs in
    there.
    """
    # TODO error_reason when user denies
    next = request.GET.get('next')
    if not next:
        next = reverse(getattr(settings, 'FACEBOOK_LOGIN_REDIRECT_URL', 'djfb_debug'))

    code = request.GET.get('code')
    if not code:
        log.error('Could not log into facebook because no code was present in '
            'the GET parameters')
        # best we can do is redirect to login page again...
        return HttpResponseRedirect(next)

    # authenticate doesn't work here, as it needs the signed request, not
    # the code. So do the validating here ourselves
    try:
        scheme = request.is_secure() and 'https' or 'http'
        redirect_uri = '%s://%s%s' % (scheme, request.get_host(), reverse('djfb_login'))
        token = request.facebook.auth.get_access_token_from_code(code,
            redirect_uri=redirect_uri)
        access_token, expires_in = token['access_token'], token['expires']
        fb_user = facebook.GraphAPI(access_token).get_object('me')
    except facebook.GraphAPIError, e:
        log.error('Could not log into facebook because: %s' % e)
        # best we can do is redirect to login page again...
        return HttpResponseRedirect(next)

    # Cache the access_token (normally autenticate does this)
    cache_access_token(fb_user['id'], access_token, expires_in)

    user = FacebookModelBackend().get_user(fb_user['id'], access_token)
    user.backend = 'django_facebook.auth.FacebookModelBackend'
    login(request, user)

    response = HttpResponseRedirect(next)
    # Set djfb_access_token and djfb_user_id, otherwise the user will be logged
    # out by our middleware
    response.set_cookie('djfb_access_token', access_token)
    response.set_cookie('djfb_user_id', fb_user['id'])

    return response


@csrf_exempt
def fb_client_login(request):
    """
    View for the POST request that is made by our javascript on login of the
    user, so the server and client are in sync concerning login status.

    The login middleware handles the login, we only need to cache the
    access_token.
    """
    if request.method != "POST":
        return HttpResponseNotAllowed(['POST'])

    if not is_fb_logged_in(request):
        return HttpResponseBadRequest('fb_client_login POSTed without valid '
                                      'fbsr_ cookie')

    cache_access_token(request.facebook.user_id,
                       request.POST['access_token'],
                       request.POST['expires_in'])
    log.debug('access_token cached in fb_client_login')
    return HttpResponse('OK')


def fb_logout(request, next=None):
    """
    Logout for the server-sided authentication flow. We can't rely on any js
    here.

    Upon logout we logout from the django auth system, we also delete the fbsr_
    cookie.

    TODO: this view does not work when you are logged in through the js SDK, due
    to the fact that the browser does not remove the fbsr_ cookie properly and
    our login middleware logs the user in again...
    """
    if next is None:
        try:
            next = settings.LOGOUT_REDIRECT_URL
        except AttributeError:
            next = reverse('djfb_debug')

    logout(request)

    response = HttpResponseRedirect(next)
    # Facebook sets the fbsr_ cookie for the "base_domain" in the fbm_ cookie
    cookie_domain = request.COOKIES.get('fbm_%s' % conf.APP_ID, '=').split('=')[1]
    response.delete_cookie(conf.COOKIE_NAME, domain=cookie_domain)
    return response



from django.conf import settings
from django.conf.urls import patterns, url
from django.views.generic import TemplateView

urlpatterns = patterns('django_facebook.views',
    url(r'^login/$', 'fb_server_login', name='djfb_login'),
    url(r'^login/client/$', 'fb_client_login', name='djfb_clientside_login'),
    url(r'^logout/$', 'fb_logout', name='djfb_logout'),
)

if settings.DEBUG:
    urlpatterns += patterns('django_facebook.views',
        url(r'^debug/$', TemplateView.as_view(
            template_name='django_facebook/debug.html'), name='djfb_debug'),
    )



from functools import update_wrapper, wraps

import facebook
from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.http import (HttpResponse, HttpResponseBadRequest,
                         HttpResponseRedirect)
from django.utils.decorators import available_attrs
from django.utils.http import urlquote

import conf
from utils import is_fb_logged_in


def canvas_only(function=None):
    """
    Decorator ensures that a page is only accessed from within a facebook
    application.

    """
    def _dec(view_func):
        def _view(request, *args, **kwargs):
            # Make sure we're receiving a signed_request from facebook
            if not request.POST.get('signed_request'):
                return HttpResponseBadRequest('<h1>400 Bad Request</h1>'
                                    '<p>Missing <em>signed_request</em>.</p>')

            # Parse the request and ensure it's valid
            signed_request = request.POST["signed_request"]
            data = conf.auth.parse_signed_request(signed_request)
            if data is False:
                return HttpResponseBadRequest('<h1>400 Bad Request</h1>'
                                  '<p>Malformed <em>signed_request</em>.</p>')

            # If the user has not authorised redirect them
            if not data.get('uid'):
                scope = getattr(settings, 'FACEBOOK_PERMS', None)
                auth_url = conf.auth.auth_url(conf.APP_ID,
                                              conf.CANVAS_PAGE,
                                              scope)
                markup = ('<script type="text/javascript">'
                          'top.location.href="%s"</script>' % auth_url)
                return HttpResponse(markup)

            # Success so return the view
            return view_func(request, *args, **kwargs)
        return _view
    return _dec(function)


def facebook_required(function=None, redirect_field_name=REDIRECT_FIELD_NAME):
    """
    Decorator for views that checks that the user is logged in, redirecting
    to the log-in page if necessary.

    """
    def _passes_test(test_func, login_url=None,
                     redirect_field_name=REDIRECT_FIELD_NAME):
        if not login_url:
            from django.conf import settings
            login_url = settings.LOGIN_URL

        def decorator(view_func):
            def _wrapped_view(request, *args, **kwargs):
                if test_func(request):
                    return view_func(request, *args, **kwargs)
                path = urlquote(request.get_full_path())
                tup = login_url, redirect_field_name, path
                return HttpResponseRedirect('%s?%s=%s' % tup)
            return wraps(view_func, assigned=available_attrs(view_func))(
                                                                _wrapped_view)
        return decorator

    actual_decorator = _passes_test(
        is_fb_logged_in,
        redirect_field_name=redirect_field_name
    )

    if function:
        return actual_decorator(function)
    return actual_decorator



import hashlib
import logging

import facebook
from django.contrib.auth import authenticate
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured

import conf
from .auth import login, logout
from .utils import (FB_DATA_CACHE_KEY, get_lazy_access_token,
                    get_signed_request_data, is_fb_logged_in)

log = logging.getLogger('django_facebook.middleware')


class FacebookAccessor(object):
    """
    Simple accessor object for the Facebook user. Non-existing properties
    will return None instead of raising a AttributeError.
    """

    def __init__(self, request):
        self.auth = conf.auth
        if is_fb_logged_in(request):
            self.user_id = request.user.get_username()
            self.access_token = get_lazy_access_token(request)
            self.graph = facebook.GraphAPI(self.access_token)

    def __getattr__(self, name):
        return None


class FacebookLoginMiddleware(object):
    """
    Transparently integrate Django accounts with Facebook. You need to add
    ``auth.FacebookModelBackend`` to ``AUTHENTICATION_BACKENDS`` for this to
    work.

    Logging in works only if ``{cookie: true}`` is passed to ``FB.init`` (by
    default when using the ``facebook_init`` template tag.

    We also allow people to log in with other backends, so we only log someone
    in if they are not already logged in.
    """
    def __init__(self, force_validate=False):
        self.force_validate = force_validate

    def process_request(self, request):
        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, 'user'):
            raise ImproperlyConfigured(
                "The FacebookCookieLoginMiddleware requires the"
                " authentication middleware to be installed. Edit your"
                " MIDDLEWARE_CLASSES setting and insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the FacebookLoginMiddleware class.")

        if request.user.is_anonymous() and conf.COOKIE_NAME in request.COOKIES:
            user = authenticate(request=request,
                                force_validate=self.force_validate)
            if user:
                login(request, user)


class FacebookLogOutMiddleware(object):
    """
    When a user logs out of facebook (on our page!), we won't get notified,
    but the fbsr_ cookies wil be cleared. So this middleware checks if that
    cookie is still present, and if not, logs the user out.

    This works only if the user is logged in with the
    ``auth.FacebookModelBackend``.
    """

    def process_request(self, request):
        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, 'user'):
            raise ImproperlyConfigured(
                "The FacebookLogOutMiddleware requires the"
                " authentication middleware to be installed. Edit your"
                " MIDDLEWARE_CLASSES setting and insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the FacebookLogOutMiddleware class.")

        if is_fb_logged_in(request):

            if not request.COOKIES.get(conf.COOKIE_NAME):
                logout(request)
                log.debug('User logged out, no fbsr_ cookie found')
                return

            data = get_signed_request_data(request)
            if data and data['user_id'] != request.user.username:
                # Also logout if the fb session changes
                logout(request)
                log.debug('User logged out. User_id on server differs from client side')


class FacebookHelperMiddleware(object):
    """
    This middleware sets the ``facebook`` attribute on the request, which
    contains the user id and a graph accessor.
    """

    def process_request(self, request):
        request.facebook = FacebookAccessor(request)


class FacebookCacheMiddleware(object):
    """
    This middleware loads tries to load user data from the cache. If found, it
    populates request.facebook with it.

    This middleware MUST come after the FacebookHelperMiddleware!
    """

    def process_request(self, request):
        if request.facebook.user_id:
            fb_data = cache.get(FB_DATA_CACHE_KEY % request.facebook.user_id)
            if fb_data:
                for k, v in fb_data.iteritems():
                    setattr(request.facebook, k, v)


class FacebookMiddleware(object):
    """
    This middleware implements the basic behaviour:

    - Log someone in if we can authenticate them (see ``auth``)
    - Log someone out if we can't authenticate them anymore
    - Add a ``facebook`` attribute to the request with a graph accessor.
    """
    def process_request(self, request):
        FacebookHelperMiddleware().process_request(request)
        FacebookLogOutMiddleware().process_request(request)
        FacebookLoginMiddleware().process_request(request)


class FacebookDebugCanvasMiddleware(object):
    """
    Emulates signed_request behaviour to test your applications embedding.

    This should be a raw string as is sent from facebook to the server in the
    POST data, obtained by LiveHeaders, Firebug or similar. This should be
    initialised before FacebookMiddleware.

    """

    def process_request(self, request):
        cp = request.POST.copy()
        request.POST = cp
        request.POST['signed_request'] = conf.DEBUG_SIGNEDREQ
        return None


class FacebookDebugCookieMiddleware(object):
    """
    Sets an imaginary cookie to make it easy to work from a development
    environment.

    This should be a raw string as is sent from a browser to the server,
    obtained by LiveHeaders, Firebug or similar. The middleware takes care of
    naming the cookie correctly. This should initialised before
    FacebookCookieLoginMiddleware.

    """

    def process_request(self, request):
        request.COOKIES[conf.COOKIE_NAME] = conf.DEBUG_COOKIE
        return None


class FacebookDebugTokenMiddleware(object):
    """
    Forces a specific access token to be used.

    This should be used instead of FacebookMiddleware. Make sure you have
    FACEBOOK_DEBUG_UID and FACEBOOK_DEBUG_TOKEN set in your configuration.

    """

    def process_request(self, request):
        request.facebook = FacebookAccessor(request)
        request.facebook.user_id = conf.DEBUG_UID
        request.facebook.access_token = conf.DEBUG_TOKEN
        return None



from setuptools import setup, find_packages

setup(
    name='django-facebook2',
    version='0.3',
    description='Facebook Authentication for Django',
    long_description=open('README.rst').read(),
    author='Tino de Bruijn',
    author_email='tinodb@gmail.com',
    url='http://github.com/tino/django-facebook2',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'django>=1.5',
        'facebook2>=2.2',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from urllib import quote

from django import template
from django.conf import settings
from django.core.urlresolvers import reverse, NoReverseMatch

from django_facebook import conf

register = template.Library()


@register.inclusion_tag('tags/facebook_load.html')
def facebook_load():
    # LANGUAGE_CODE defaults to en-us, which is also the default for FB
    lang, country = settings.LANGUAGE_CODE.split('-')
    fb_locale = "%s_%s" % (lang, country.upper())
    return {'fb_locale': fb_locale}


@register.tag
def facebook_init(parser, token):
    nodelist = parser.parse(('endfacebook_init',))
    parser.delete_first_token()
    return FacebookNode(nodelist)


class FacebookNode(template.Node):
    """ Allow code to be added inside the facebook asynchronous closure. """
    def __init__(self, nodelist):
        try:
            app_id = conf.APP_ID
        except AttributeError:
            raise template.TemplateSyntaxError, "%r tag requires " \
                "FACEBOOK_APP_ID to be configured." \
                % token.contents.split()[0]
        self.app_id = app_id
        self.nodelist = nodelist

    def render(self, context):
        t = template.loader.get_template('tags/facebook_init.html')
        code = self.nodelist.render(context)
        custom_context = context
        custom_context['code'] = code
        custom_context['app_id'] = self.app_id
        custom_context['version'] = getattr(settings, 'FACEBOOK_VERSION', '2.2')
        return t.render(custom_context)


@register.simple_tag
def facebook_perms():
    return ",".join(getattr(settings, 'FACEBOOK_PERMS', []))


@register.simple_tag
def facebook_login_url(request, redirect_after=None):
    """
    Templatetag for rendering the login url for server-sided login.

    An extra redirect url or urlname can be passed to be redirected to after
    server-side login has happened.
    """
    login_url = 'https://www.facebook.com/dialog/oauth?' + \
        'client_id=%s&scope=%s&redirect_uri=%s'
    redirect_to = reverse('djfb_login')
    if redirect_after:
        if redirect_after.startswith('/'):
            redirect_to += '?next=%s' % quote(redirect_after)
        else:
            redirect_to += '?next=%s' % quote(reverse(redirect_after))
    return login_url % (conf.APP_ID,
                        facebook_perms(),
                        request.build_absolute_uri(redirect_to))



import facebook
from celery.utils.log import get_task_logger
from django.core.exceptions import ImproperlyConfigured

from .utils import get_cached_access_token

try:
    from celery import shared_task, subtask
except ImportError:
    raise ImproperlyConfigured('You need to install a recent version of celery'
                               ' to use the tasks!')

log = get_task_logger(__name__)


@shared_task(bind=True, default_retry_delay=60)
def get_friends_for_user(self, fb_id, callback, next_uri=None):
    """
    Get the facebook friends for the user with fb_id.

    1. Needs a valid access_token in the cache
    2. Needs 'user_friends' permission
    3. Needs a callback function that can store the friends somewhere

    If 1. is not present, the task is delayed
    If 2. is not the case, you're out of luck
    """
    access_token = get_cached_access_token(fb_id)
    if access_token is None:
        raise self.retry(exc=ValueError("Failed to fetch facebook data for %s. "
                                        "No access_token found in cache" % fb_id))

    graph = facebook.GraphAPI(access_token)

    try:
        if next_uri:
            data = graph.bare_request(next_uri)
        else:
            data = graph.get_connections('me', 'friends', limit=500)
    except facebook.GraphAPIError as exc:
        raise self.retry(exc=exc)

    subtask(callback).delay(data['data'])
    if data['paging'].get('next'):
        self.delay(fb_id, callback, next_uri=data['paging']['next'])
